#include <stdio.h>
#include <mpi.h>
#include <malloc.h>

int main (int nargs, char** args)
{
  int size, my_rank, i, m = 1000, lm, lower_bound, res;
  double l_sum, g_sum, time, x, dx, *u, *v;

  MPI_Init (&nargs, &args);
  MPI_Comm_size (MPI_COMM_WORLD, &size);
  MPI_Comm_rank (MPI_COMM_WORLD, &my_rank);

  if (my_rank==0 && nargs>1)
    m = atoi(args[1]);

  MPI_Bcast (&m, 1, MPI_INT, 0, MPI_COMM_WORLD);
  dx = 1.0/m;

  time = MPI_Wtime();

  /* data partition & load balancing */
  lm = m/size;
  res = m%size;
  lower_bound = my_rank*lm;
  lower_bound += (my_rank<=res) ? my_rank : res;
  if (my_rank+1<=res)
    ++lm;

  /* allocation of data storage */
  u = (double*) malloc (lm*sizeof(double));
  v = (double*) malloc (lm*sizeof(double));

  /* fill out the u og v-vectors */
  x = lower_bound*dx;
  for (i=0; i<lm; i++) {
    x += dx;
    u[i] = x;
    v[i] = x*x;
  }

  /* calculate the local result */
  l_sum = 0.;
  for (i=0; i<lm; i++)
    l_sum += u[i]*v[i];

  MPI_Allreduce (&l_sum,&g_sum,1,MPI_DOUBLE,
                 MPI_SUM,MPI_COMM_WORLD);

  time = MPI_Wtime()-time;

  /* output the global result */
  printf("<%d>m=%d, lower_bound=%d, lm=%d, u*v=%g, time=%g\n",
         my_rank, m, lower_bound, lm, g_sum*dx, time);

  free (u);
  free (v);

  MPI_Finalize ();
  return 0;
}   
