#include <stdio.h>
#include <mpi.h>
#include <math.h>

extern void jacobi_iteration (int my_rank, int size, int n_i,
			      double* rhs, double* x_k1, double* x_k);
extern void calc_residual (int my_rank, int size, int n_i,
                           double* rhs, double* x, double* res);
extern double norm (int n_i, double* x);

int main (int nargs, char** args)
{
  int size, my_rank, i, j, n = 99999, n_i, lower_bound;
  double time, base, r_norm, x, dx, *rhs, *res, *x_k, *x_k1;

  MPI_Init (&nargs, &args);
  MPI_Comm_size (MPI_COMM_WORLD, &size);
  MPI_Comm_rank (MPI_COMM_WORLD, &my_rank);

  if (my_rank==0 && nargs>1)
    n = atoi(args[1]);

  MPI_Bcast (&n, 1, MPI_INT, 0, MPI_COMM_WORLD);
  dx = 1.0/(n+1);

  time = MPI_Wtime();

  /* data partition & load balancing */
  n_i = n/size;
  i = n%size;
  lower_bound = my_rank*n_i;
  lower_bound += (my_rank<=i) ? my_rank : i;
  if (my_rank+1<=i)
    ++n_i;

  /* allocation of data storage, plus 2 (ghost) boundary points */
  x_k = (double*) malloc ((n_i+2)*sizeof(double));
  x_k1= (double*) malloc ((n_i+2)*sizeof(double));
  rhs = (double*) malloc ((n_i+2)*sizeof(double));
  res = (double*) malloc ((n_i+2)*sizeof(double));

  /* fill out the rhs og x_k-vectors */
  x = lower_bound*dx;
  for (i=1; i<=n_i; i++) {
    x += dx;
    rhs[i] = dx*dx*M_PI*M_PI*sin(M_PI*x);
    x_k[i] = x_k1[i] = 0.;
  }
  x_k[0]=x_k[n_i+1]=x_k1[0]=x_k1[n_i+1]=0.;

  base = norm(n_i,rhs);
  r_norm = 1.0;
  i = 0;
  while (i<1000 && (r_norm/base)>1.0e-4) {
    ++i;
    jacobi_iteration(my_rank,size,n_i,rhs,x_k1,x_k);
    calc_residual(my_rank,size,n_i,rhs,x_k,res);
    r_norm = norm(n_i,res);
    for (j=0; j<n_i+2; j++)
      x_k1[j] = x_k[j];
  }

  time = MPI_Wtime()-time;

  printf("<%d> n+1=%d, iters=%d, time=%g\n",my_rank,n+1,i,time);

  free (x_k); free (x_k1); free (rhs); free (res);

  MPI_Finalize ();
  return 0;
}   
